@@ -180,7 +180,7 @@ main(int argc, char *argv[]) {
 				pseudo_diag("Can't resolve path '%s'\n", optarg);
 				usage(EXIT_FAILURE);
 			}
-			opt_i = s;
+			opt_i = strdup(s);
 			break;
 		case 'l': /* log */
 			optptr += snprintf(optptr, pseudo_path_max() - (optptr - opts),
@@ -193,7 +193,7 @@ main(int argc, char *argv[]) {
 				pseudo_diag("Can't resolve move-from path '%s'\n", optarg);
 				usage(EXIT_FAILURE);
 			}
-			opt_m = s;
+			opt_m = strdup(s);
 			break;
 		case 'M': /* move to... (see also 'm') */
 			s = PSEUDO_ROOT_PATH(AT_FDCWD, optarg, 0);
@@ -201,7 +201,7 @@ main(int argc, char *argv[]) {
 				pseudo_diag("Can't resolve move-to path '%s'\n", optarg);
 				usage(EXIT_FAILURE);
 			}
-			opt_M = s;
+			opt_M = strdup(s);
 			break;
 		case 'p': /* passwd file path */
 			s = PSEUDO_ROOT_PATH(AT_FDCWD, optarg, AT_SYMLINK_NOFOLLOW);
@@ -228,7 +228,7 @@ main(int argc, char *argv[]) {
 			}
 			pseudo_set_value("PSEUDO_CHROOT", s);
 			if (o == 'r')
-				opt_r = s;
+				opt_r = strdup(s);
 			break;
 		case 'S': /* stop */
 			opt_S = 1;
@@ -343,7 +343,8 @@ main(int argc, char *argv[]) {
 		return pseudo_db_check(opt_B);
 	}
 
-	if (opt_S) {
+        /* If you didn't specify a command, opt_S shuts down here. */
+	if (opt_S && argc <= optind) {
 		return pseudo_client_shutdown();
 	}
 
@@ -418,8 +419,17 @@ main(int argc, char *argv[]) {
 			/* try to hint that we don't think we still need
 			 * the server.
 			 */
-			pseudo_client_shutdown();
-			return WEXITSTATUS(rc);
+                        if (opt_S) {
+				pseudo_client_shutdown();
+                        }
+			if (WIFEXITED(rc)) {
+				return WEXITSTATUS(rc);
+			} else if (WIFSIGNALED(rc)) {
+				kill(getpid(), WTERMSIG(rc));
+				exit(1);
+			} else {
+				exit(1);
+			}
 		} else {
 			rc = execv(fullpath, argv);
 			if (rc == -1) {
@@ -497,7 +507,7 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 	int found_path = 0, found_ino = 0;
 	int prefer_ino = 0;
 	int xattr_flags = 0;
-	int trailing_slash = 0, old_trailing_slash = 0;
+	int trailing_slash = 0;
 
 	if (!msg)
 		return 1;
@@ -548,7 +558,6 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 			if (oldpathlen > 0 && msg->op == OP_RENAME) {
 				if (oldpath[oldpathlen - 1] == '/') {
 					oldpath[--oldpathlen] = '\0';
-					old_trailing_slash = 1;
 				}
 			}
 			/* if we got an oldpath, but a 0-length initial
@@ -824,6 +833,8 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 			pdb_unlink_file_dev(&by_ino);
 		}
 		if (!found_path) {
+			pseudo_debug(PDBGF_DB, "linking %s for OP_CREAT\n",
+				msg->pathlen ? msg->path : "no path");
 			pdb_link_file(msg, NULL);
 		} else {
 			/* again, an error, but leaving it alone for now. */
@@ -857,6 +868,8 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 		 */
 		if (!found_path && !found_ino) {
 			pseudo_debug(PDBGF_FILE, "(new) ");
+			pseudo_debug(PDBGF_DB, "linking %s for OP_[F]CHMOD\n",
+				msg->pathlen ? msg->path : "no path");
 			pdb_link_file(msg, NULL);
 		}
 		break;
@@ -887,6 +900,8 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 		 */
 		if (!found_path && !found_ino) {
 			pseudo_debug(PDBGF_FILE, "(new) ");
+			pseudo_debug(PDBGF_DB, "linking %s for OP_[F]CHOWN\n",
+				msg->pathlen ? msg->path : "no path");
 			pdb_link_file(msg, NULL);
 		}
 		break;
@@ -896,7 +911,19 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 		 * case where there might be a clash.
 		 */
 		if (found_ino || found_path) {
-			*msg = db_header;
+			#ifdef PSEUDO_XATTRDB
+			if (db_header.uid == (uid_t) -1 &&
+			    db_header.gid == (gid_t) -1) {
+				/* special case: this row was created
+				 * to allow xattr lookups, and it's not
+				 * actually valid data.
+				 */
+				msg->result = RESULT_FAIL;
+			} else
+			#endif
+			{
+				*msg = db_header;
+			}
 		} else {
 			msg->result = RESULT_FAIL;
 		}
@@ -942,6 +969,9 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 		} else {
 			*msg = msg_header;
 		}
+		pseudo_debug(PDBGF_DB, "linking %s for %s\n",
+			msg->pathlen ? msg->path : "no path",
+			pseudo_op_name(msg->op));
 		pdb_link_file(msg, NULL);
 		break;
 	case OP_RENAME:
@@ -1004,6 +1034,9 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 			pdb_unlink_file_dev(&by_ino);
 		}
 		*msg = msg_header;
+		pseudo_debug(PDBGF_DB, "linking %s for %s\n",
+			msg->pathlen ? msg->path : "no path",
+			pseudo_op_name(msg->op));
 		pdb_link_file(msg, NULL);
 		break;
 	case OP_GET_XATTR:
@@ -1036,6 +1069,18 @@ pseudo_op(pseudo_msg_t *msg, const char *program, const char *tag, char **respon
 	case OP_SET_XATTR:
 		/* we need a row entry to store xattr info */
 		if (row == -1) {
+#ifdef PSEUDO_XATTRDB
+			/* mark the entry as Not Reliable for purposes
+			 * of stat-type calls, since changes wouldn't
+			 * get reported to us.
+			 */
+			msg->uid = (uid_t) -1;
+			msg->gid = (gid_t) -1;
+#endif
+			pseudo_debug(PDBGF_DB | PDBGF_XATTR, "linking %s (uid -1 if xattr) for %s, mode 0%o\n",
+				msg->pathlen ? msg->path : "no path",
+				pseudo_op_name(msg->op),
+                                (int) msg->mode);
 			pdb_link_file(msg, &row);
 		}
 		if (pdb_set_xattr(row, oldpath, oldpathlen, xattr_flags)) {
@@ -1069,7 +1114,11 @@ int
 pseudo_server_response(pseudo_msg_t *msg, const char *program, const char *tag, char **response_path, size_t *response_len) {
 	switch (msg->type) {
 	case PSEUDO_MSG_PING:
-		msg->result = RESULT_SUCCEED;
+                /* mad hackery: if we aren't logging, the client gets told
+                 * not to send open/exec notifications, which have no other
+                 * purpose.
+                 */
+		msg->result = opt_l ? RESULT_SUCCEED : RESULT_FAIL;
 		if (opt_l)
 			pdb_log_msg(SEVERITY_INFO, msg, program, tag, NULL);
 		return 0;
