@@ -25,6 +25,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <limits.h>
+#include <time.h>
 #include <unistd.h>
 #include <errno.h>
 #include <sys/socket.h>
@@ -37,6 +38,10 @@
 #include <pwd.h>
 #include <grp.h>
 
+#ifdef PSEUDO_XATTRDB
+#include <sys/xattr.h>
+#endif
+
 #include "pseudo.h"
 #include "pseudo_ipc.h"
 #include "pseudo_client.h"
@@ -70,6 +75,7 @@ char *pseudo_cwd_rel = NULL;
 int pseudo_disabled = 0;
 int pseudo_allow_fsync = 0;
 static int pseudo_local_only = 0;
+static int pseudo_client_logging = 1;
 
 int pseudo_umask = 022;
 
@@ -77,8 +83,12 @@ static char **fd_paths = NULL;
 static int nfds = 0;
 static const char **passwd_paths = NULL;
 static int npasswd_paths = 0;
-static int messages = 0;
-static struct timeval message_time = { .tv_sec = 0 };
+#ifdef PSEUDO_PROFILING
+int pseudo_profile_fd = -1;
+static int profile_interval = 1;
+static pseudo_profile_t profile_data;
+struct timeval *pseudo_wrapper_time = &profile_data.wrapper_time;
+#endif
 static int pseudo_inited = 0;
 
 static int sent_messages = 0;
@@ -186,6 +196,244 @@ build_passwd_paths(void)
 	return;
 }
 
+#ifdef PSEUDO_XATTRDB
+/* We really want to avoid calling the wrappers for these inside the
+ * implementation. pseudo_wrappers will reinitialize these after it's
+ * gotten the real_* found.
+ */
+ssize_t (*pseudo_real_lgetxattr)(const char *, const char *, void *, size_t) = lgetxattr;
+ssize_t (*pseudo_real_fgetxattr)(int, const char *, void *, size_t) = fgetxattr;
+int (*pseudo_real_lsetxattr)(const char *, const char *, const void *, size_t, int) = lsetxattr;
+int (*pseudo_real_fsetxattr)(int, const char *, const void *, size_t, int) = fsetxattr;
+/* Executive summary: We use an extended attribute,
+ * user.pseudo_data, to store exactly the data we would otherwise
+ * have stored in the database. Which is to say, uid, gid, mode, rdev.
+ *
+ * If we don't find a value, save an empty one with a lower version
+ * number to indicate that we don't have data to reduce round trips.
+ */
+typedef struct {
+	int version;
+	uid_t uid;
+	gid_t gid;
+	mode_t mode;
+	dev_t rdev;
+} pseudo_db_data_t;
+
+static pseudo_msg_t xattrdb_data;
+
+pseudo_msg_t *
+pseudo_xattrdb_save(int fd, const char *path, const struct stat64 *buf) {
+	int rc = -1;
+	if (!path && fd < 0)
+		return NULL;
+	if (!buf)
+		return NULL;
+	pseudo_db_data_t pseudo_db_data = {
+		.version = 1,
+		.uid = buf->st_uid,
+		.gid = buf->st_gid,
+		.mode = buf->st_mode,
+		.rdev = buf->st_rdev
+	};
+	if (path) {
+		rc = pseudo_real_lsetxattr(path, "user.pseudo_data", &pseudo_db_data, sizeof(pseudo_db_data), 0);
+	} else if (fd >= 0) {
+		rc = pseudo_real_fsetxattr(fd, "user.pseudo_data", &pseudo_db_data, sizeof(pseudo_db_data), 0);
+	}
+	pseudo_debug(PDBGF_XATTRDB, "tried to save data for %s/%d: uid %d, mode %o, rc %d.\n",
+		path ? path : "<nil>", fd, (int) pseudo_db_data.uid, (int) pseudo_db_data.mode, rc);
+	/* none of the other fields are checked on save, and the value
+	 * is currently only really used by mknod.
+	 */
+	if (rc == 0) {
+		xattrdb_data.result = RESULT_SUCCEED;
+		return &xattrdb_data;
+	}
+	return NULL;
+}
+
+pseudo_msg_t *
+pseudo_xattrdb_load(int fd, const char *path, const struct stat64 *buf) {
+	int rc = -1, retryrc = -1;
+	if (!path && fd < 0)
+		return NULL;
+	/* don't try to getxattr on a thing unless we think it is
+	 * likely to work.
+	 */
+	if (buf) {
+		if (!S_ISDIR(buf->st_mode) && !S_ISREG(buf->st_mode)) {
+			return NULL;
+		}
+	}
+	pseudo_db_data_t pseudo_db_data;
+	if (path) {
+		rc = pseudo_real_lgetxattr(path, "user.pseudo_data", &pseudo_db_data, sizeof(pseudo_db_data));
+		if (rc == -1) {
+			pseudo_db_data = (pseudo_db_data_t) { .version = 0 };
+			retryrc = pseudo_real_lsetxattr(path, "user.pseudo_data", &pseudo_db_data, sizeof(pseudo_db_data), 0);
+		}
+	} else if (fd >= 0) {
+		rc = pseudo_real_fgetxattr(fd, "user.pseudo_data", &pseudo_db_data, sizeof(pseudo_db_data));
+		if (rc == -1) {
+			pseudo_db_data = (pseudo_db_data_t) { .version = 0 };
+			retryrc = pseudo_real_fsetxattr(fd, "user.pseudo_data", &pseudo_db_data, sizeof(pseudo_db_data), 0);
+		}
+	}
+	pseudo_debug(PDBGF_XATTRDB, "tried to load data for %s[%d]: rc %d, version %d.\n",
+		path ? path : "<nil>", fd, rc, pseudo_db_data.version);
+	if (rc == -1 && retryrc == 0) {
+		/* there's no data, but there could have been; treat
+		 * this as an empty database result.
+		 */
+		pseudo_debug(PDBGF_XATTRDB, "wrote version 0 for %s[%d]\n",
+			path ? path : "<nil>", fd);
+		xattrdb_data.result = RESULT_FAIL;
+		return &xattrdb_data;
+	} else if (rc == -1) {
+		/* we can't create an extended attribute, so we may have
+		 * used the database.
+		 */
+		return NULL;
+	}
+	/* Version 0 = just recording that we looked and found
+	 * nothing.
+	 * Version 1 = actually implemented.
+	 */
+	switch (pseudo_db_data.version) {
+	case 0:
+	default:
+		xattrdb_data.result = RESULT_FAIL;
+		break;
+	case 1:
+		xattrdb_data.uid = pseudo_db_data.uid;
+		xattrdb_data.gid = pseudo_db_data.gid;
+		xattrdb_data.mode = pseudo_db_data.mode;
+		xattrdb_data.rdev = pseudo_db_data.rdev;
+		xattrdb_data.result = RESULT_SUCCEED;
+		break;
+	}
+	return &xattrdb_data;
+}
+#endif
+
+#ifdef PSEUDO_PROFILING
+static int pseudo_profile_pid = -2;
+
+static void
+pseudo_profile_start(void) {
+	/* We use -1 as a starting value, and -2 as a value
+	 * indicating not to try to open it.
+	 */
+	int existing_data = 0;
+	int pid = getpid();
+	if (pseudo_profile_pid > 0 && pseudo_profile_pid != pid) {
+		/* looks like there's been a fork. We intentionally
+		 * abandon any existing stats, since the parent might
+		 * want to write them, and we want to combine with
+		 * any previous stats for this pid.
+		 */
+		close(pseudo_profile_fd);
+		pseudo_profile_fd = -1;
+	}
+	if (pseudo_profile_fd == -1) {
+		int fd = -2;
+		char *profile_path = pseudo_get_value("PSEUDO_PROFILE_PATH");
+		pseudo_profile_pid = pid;
+		if (profile_path) {
+			fd = open(profile_path, O_RDWR | O_CREAT, 0600);
+			if (fd >= 0) {
+				fd = pseudo_fd(fd, MOVE_FD);
+			}
+			if (fd < 0) {
+				fd = -2;
+			} else {
+				if (pid > 0) {
+					pseudo_profile_t data;
+					off_t rc;
+
+					rc = lseek(fd, pid * sizeof(data), SEEK_SET);
+					if (rc == (off_t) (pid * sizeof(data))) {
+						rc = read(fd, &profile_data, sizeof(profile_data));
+						/* cumulative with other values in same file */
+						if (rc == sizeof(profile_data)) {
+							pseudo_debug(PDBGF_PROFILE, "pid %d found existing profiling data.\n", pid);
+							existing_data = 1;
+							++profile_data.processes;
+						} else {
+							pseudo_debug(PDBGF_PROFILE, "read failed for pid %d: %d, %d\n", pid, (int) rc, errno);
+						}
+						profile_interval = 1;
+					}
+				}
+			}
+		}
+		pseudo_profile_fd = fd;
+	} else {
+		pseudo_debug(PDBGF_PROFILE, "_start called with existing fd? (pid %d)", (int) pseudo_profile_pid);
+		existing_data = 1;
+		++profile_data.processes;
+		profile_data.total_ops = 0;
+		profile_data.messages = 0;
+		profile_data.wrapper_time = (struct timeval) { .tv_sec = 0 };
+		profile_data.op_time = (struct timeval) { .tv_sec = 0 };
+		profile_data.ipc_time = (struct timeval) { .tv_sec = 0 };
+	}
+	if (!existing_data) {
+		pseudo_debug(PDBGF_PROFILE, "pid %d found no existing profiling data.\n", pid);
+		profile_data = (pseudo_profile_t) {
+			.processes = 1,
+			.total_ops = 0,
+			.messages = 0,
+			.wrapper_time = (struct timeval) { .tv_sec = 0 },
+			.op_time = (struct timeval) { .tv_sec = 0 },
+			.ipc_time = (struct timeval) { .tv_sec = 0 },
+		};
+	}
+}
+
+static inline void
+fix_tv(struct timeval *tv) {
+	if (tv->tv_usec > 1000000) {
+		tv->tv_sec += tv->tv_usec / 1000000;
+		tv->tv_usec %= 1000000;
+	} else if (tv->tv_usec < 0) {
+		/* C99 and later guarantee truncate-towards-zero.
+		 * We want -1 through -1000000 usec to produce
+		 * -1 seconds, etcetera. Note that sec is
+		 * negative, so yes, we want to add to tv_sec and
+		 * subtract from tv_usec.
+		 */
+		int sec = (tv->tv_usec - 999999) / 1000000;
+		tv->tv_sec += sec;
+		tv->tv_usec -= 1000000 * sec;
+	}
+}
+
+static int profile_reported = 0;
+static void
+pseudo_profile_report(void) {
+	if (pseudo_profile_fd < 0) {
+		return;
+	}
+	fix_tv(&profile_data.wrapper_time);
+	fix_tv(&profile_data.op_time);
+	fix_tv(&profile_data.ipc_time);
+	if (pseudo_profile_pid >= 0) {
+		int rc1, rc2;
+		rc1 = lseek(pseudo_profile_fd, pseudo_profile_pid * sizeof(profile_data), SEEK_SET);
+		rc2 = write(pseudo_profile_fd, &profile_data, sizeof(profile_data));
+		if (!profile_reported) {
+			pseudo_debug(PDBGF_PROFILE, "pid %d (%s) saving profiling info: %d, %d.\n",
+				pseudo_profile_pid,
+				program_invocation_name ? program_invocation_name : "unknown",
+				rc1, rc2);
+			profile_reported = 1;
+		}
+	}
+}
+#endif
+
 void
 pseudo_init_client(void) {
 	char *env;
@@ -196,6 +444,12 @@ pseudo_init_client(void) {
 		close(connect_fd);
 		connect_fd = -1;
 	}
+#ifdef PSEUDO_PROFILING
+	if (pseudo_profile_fd > -1) {
+		close(pseudo_profile_fd);
+	}
+	pseudo_profile_fd = -1;
+#endif
 
 	/* in child processes, PSEUDO_DISABLED may have become set to
 	 * some truthy value, in which case we'd disable pseudo,
@@ -268,7 +522,7 @@ pseudo_init_client(void) {
 	env = getenv("PSEUDO_UNLOAD");
 	if (env) {
 		pseudo_set_value("PSEUDO_UNLOAD", env);
-		pseudo_disabled=1;
+		pseudo_disabled = 1;
 	}
 
 	/* Setup global items needed for pseudo to function... */
@@ -403,8 +657,12 @@ pseudo_init_client(void) {
 
 		pseudo_inited = 1;
 	}
-	if (!pseudo_disabled)
+	if (!pseudo_disabled) {
 		pseudo_client_getcwd();
+#ifdef PSEUDO_PROFILING
+		pseudo_profile_start();
+#endif
+	}
 
 	pseudo_magic();
 }
@@ -810,7 +1068,14 @@ client_ping(void) {
 		/* and that's not good, so... */
 		server_pid = 0;
 		return 1;
-	}
+	} else {
+                /* The server tells us whether or not to log things. */
+                if (ack->result == RESULT_SUCCEED) {
+                        pseudo_client_logging = 1;
+                } else {
+                        pseudo_client_logging = 0;
+                }
+        }
 	pseudo_debug(PDBGF_CLIENT | PDBGF_VERBOSE, "ping ok\n");
 	return 0;
 }
@@ -936,7 +1201,8 @@ pseudo_client_setup(void) {
 	}
 	pseudo_debug(PDBGF_CLIENT, "server seems to be gone, trying to restart\n");
 	if (client_spawn_server()) {
-		pseudo_debug(PDBGF_CLIENT, "failed to spawn server, giving up.\n");
+
+		pseudo_debug(PDBGF_CLIENT, "failed to spawn server, waiting for retry.\n");
 		return 1;
 	} else {
 		pseudo_debug(PDBGF_CLIENT, "restarted, new pid %d\n", server_pid);
@@ -944,10 +1210,11 @@ pseudo_client_setup(void) {
 			return 0;
 		}
 	}
-	pseudo_debug(PDBGF_CLIENT, "couldn't get a server, giving up.\n");
+	pseudo_debug(PDBGF_CLIENT, "couldn't get or spawn a server.\n");
 	return 1;
 }
 
+#define PSEUDO_RETRIES 50
 static pseudo_msg_t *
 pseudo_client_request(pseudo_msg_t *msg, size_t len, const char *path) {
 	pseudo_msg_t *response = 0;
@@ -957,51 +1224,55 @@ pseudo_client_request(pseudo_msg_t *msg, size_t len, const char *path) {
 	if (!msg)
 		return 0;
 
-	do {
-		do {
-			pseudo_debug(PDBGF_CLIENT | PDBGF_VERBOSE, "sending a message: ino %llu\n",
-				(unsigned long long) msg->ino);
-			if (connect_fd < 0) {
-				pseudo_debug(PDBGF_CLIENT, "trying to get server\n");
-				if (pseudo_client_setup()) {
-					return 0;
-				}
+	/* Try to send a message. If sending fails, try to spawn a server,
+	 * and whether or not we succeed, wait a little bit and retry sending.
+	 * It's okay if we can't start a server sometimes, because another
+	 * client may have done it.
+	 */
+        for (tries = 0; tries < PSEUDO_RETRIES; ++tries) {
+		pseudo_debug(PDBGF_CLIENT | PDBGF_VERBOSE, "sending a message: ino %llu\n",
+			(unsigned long long) msg->ino);
+		rc = pseudo_msg_send(connect_fd, msg, len, path);
+		if (rc != 0) {
+			pseudo_debug(PDBGF_CLIENT | PDBGF_VERBOSE, "msg_send: %d%s\n",
+				rc,
+				rc == -1 ? " (sigpipe)" :
+					   " (short write)");
+			pseudo_debug(PDBGF_CLIENT, "trying to get server, try %d\n", tries);
+			/* try to open server; if we fail, wait a bit before
+			 * retry.
+			 */
+			if (pseudo_client_setup()) {
+				int ms = (getpid() % 5) + 3 + tries;
+				struct timespec delay = { .tv_sec = 0, .tv_nsec = ms * 1000000 };
+				nanosleep(&delay, NULL);
 			}
-			rc = pseudo_msg_send(connect_fd, msg, len, path);
-			if (rc != 0) {
-				pseudo_debug(PDBGF_CLIENT | PDBGF_VERBOSE, "msg_send: %d%s\n",
-					rc,
-					rc == -1 ? " (sigpipe)" :
-					           " (short write)");
-				pseudo_client_setup();
-				++tries;
-				if (tries > 3) {
-					pseudo_debug(PDBGF_CLIENT, "Can't get server going again.\n");
-					return 0;
-				}
-			}
-		} while (rc != 0);
+			continue;
+		}
+		/* note "continue" above; we only get here if rc was 0,
+		 * indicating a successful send.
+		 */
 		pseudo_debug(PDBGF_CLIENT | PDBGF_VERBOSE, "sent!\n");
-                if (msg->type != PSEUDO_MSG_FASTOP) {
-                        response = pseudo_msg_receive(connect_fd);
-                        if (!response) {
-                                ++tries;
-                                if (tries > 3) {
-                                        pseudo_debug(PDBGF_CLIENT, "Can't get responses.\n");
-                                        return 0;
-                                }
-                        }
-                } else {
-                        return 0;
-                }
-	} while (response == 0);
-	if (response->type != PSEUDO_MSG_ACK) {
-		pseudo_debug(PDBGF_CLIENT, "got non-ack response %d\n", response->type);
-		return 0;
-	} else {
-		pseudo_debug(PDBGF_CLIENT | PDBGF_VERBOSE, "got response type %d\n", response->type);
+		if (msg->type != PSEUDO_MSG_FASTOP) {
+			response = pseudo_msg_receive(connect_fd);
+			if (!response) {
+				pseudo_debug(PDBGF_CLIENT, "expected response did not occur; retrying\n");
+			} else {
+				if (response->type != PSEUDO_MSG_ACK) {
+					pseudo_debug(PDBGF_CLIENT, "got non-ack response %d\n", response->type);
+					return 0;
+				} else {
+					pseudo_debug(PDBGF_CLIENT | PDBGF_VERBOSE, "got response type %d\n", response->type);
+					return response;
+				}
+			}
+		} else {
+			return 0;
+		}
 	}
-	return response;
+	pseudo_diag("pseudo: server connection persistently failed, aborting.\n");
+	abort();
+	return 0;
 }
 
 int
@@ -1095,7 +1366,7 @@ base_path(int dirfd, const char *path, int leave_last) {
 	if (!path)
 		return NULL;
 	if (!*path)
-		return strdup("");
+		return "";
 
 	if (path[0] != '/') {
 		if (dirfd != -1 && dirfd != AT_FDCWD) {
@@ -1152,14 +1423,106 @@ pseudo_client_op(pseudo_op_t op, int access, int fd, int dirfd, const char *path
 	static size_t alloced_len = 0;
 	int strip_slash;
 
+#ifdef PSEUDO_PROFILING
+	struct timeval tv1_op, tv2_op;
+
+	gettimeofday(&tv1_op, NULL);
+	++profile_data.total_ops;
+#endif
 	/* disable wrappers */
 	pseudo_antimagic();
 
+	/* note: I am not entirely sure this should happen even if no
+	 * messages have actually been sent. */
 	if (!sent_messages) {
 		sent_messages = 1;
 		atexit(void_client_ping);
 	}
 
+	/* if path isn't available, try to find one? */
+	if (!path && fd >= 0 && fd <= nfds) {
+		path = fd_path(fd);
+		if (!path) {
+			pathlen = 0;
+		} else {
+			pathlen = strlen(path) + 1;
+		}
+	}
+
+#ifdef PSEUDO_XATTRDB
+	if (buf) {
+		struct stat64 bufcopy = *buf;
+		int do_save = 0;
+		/* maybe use xattr instead */
+		/* note: if we use xattr, logging won't work reliably
+		 * because the server won't get messages if these work.
+		 */
+		switch (op) {
+		case OP_CHMOD:
+		case OP_FCHMOD:
+		case OP_CHOWN:
+		case OP_FCHOWN:
+			/* for these, we want to start with the existing db
+			 * values.
+			 */
+			bufcopy = *buf;
+			result = pseudo_xattrdb_load(fd, path, buf);
+			if (result && result->result == RESULT_SUCCEED) {
+				pseudo_debug(PDBGF_XATTR, "merging existing values for xattr\n");
+				switch (op) {
+				case OP_CHMOD:
+				case OP_FCHMOD:
+					bufcopy.st_uid = result->uid;
+					bufcopy.st_gid = result->gid;
+					break;
+				case OP_CHOWN:
+				case OP_FCHOWN:
+					bufcopy.st_rdev = result->rdev;
+					bufcopy.st_mode = result->mode;
+					break;
+				default:
+					break;
+				}
+				
+			} else {
+				switch (op) {
+				case OP_CHMOD:
+				case OP_FCHMOD:
+					bufcopy.st_uid = pseudo_fuid;
+					bufcopy.st_gid = pseudo_fgid;
+					break;
+				default:
+					break;
+				}
+			}
+			result = NULL;
+			do_save = 1;
+			break;
+		case OP_CREAT:
+		case OP_MKDIR:
+		case OP_MKNOD:
+			bufcopy.st_uid = pseudo_fuid;
+			bufcopy.st_gid = pseudo_fgid;
+			do_save = 1;
+			break;
+		case OP_LINK:
+			do_save = 1;
+			break;
+		case OP_FSTAT:
+		case OP_STAT:
+			result = pseudo_xattrdb_load(fd, path, buf);
+			break;
+		default:
+			break;
+		}
+		if (do_save) {
+			result = pseudo_xattrdb_save(fd, path, &bufcopy);
+		}
+		if (result)
+			goto skip_path;
+	}
+#endif
+
 	if (op == OP_RENAME) {
 		va_list ap;
 		if (!path) {
@@ -1208,16 +1571,6 @@ pseudo_client_op(pseudo_op_t op, int access, int fd, int dirfd, const char *path
 			op == OP_GET_XATTR ? "get" : "remove", path_extra_1);
 	}
 
-	/* if path isn't available, try to find one? */
-	if (!path && fd >= 0 && fd <= nfds) {
-		path = fd_path(fd);
-		if (!path) {
-			pathlen = 0;
-		} else {
-			pathlen = strlen(path) + 1;
-		}
-	}
-
 	if (path) {
 		if (pathlen == (size_t) -1) {
 			pathlen = strlen(path) + 1;
@@ -1288,30 +1641,41 @@ pseudo_client_op(pseudo_op_t op, int access, int fd, int dirfd, const char *path
 		}
 	}
 
-	pseudo_debug(PDBGF_OP, "%s%s", pseudo_op_name(op),
-		(dirfd != -1 && dirfd != AT_FDCWD && op != OP_DUP) ? "at" : "");
-	if (path_extra_1) {
-		pseudo_debug(PDBGF_OP, " %s ->", path_extra_1);
-	}
-	if (path) {
-		pseudo_debug(PDBGF_OP, " %s", path);
-	}
-	/* for OP_RENAME in renameat, "fd" is also used for the
-	 * second dirfd.
+#ifdef PSEUDO_XATTRDB
+	/* If we were able to store things in xattr, we can easily skip
+	 * most of the fancy path computations and such.
 	 */
-	if (fd != -1 && op != OP_RENAME) {
-		pseudo_debug(PDBGF_OP, " [fd %d]", fd);
-	}
-	if (buf) {
-		pseudo_debug(PDBGF_OP, " (+buf)");
+	skip_path:
+#endif
+
+	if (buf)
 		pseudo_msg_stat(&msg, buf);
-		if (buf && fd != -1) {
-			pseudo_debug(PDBGF_OP, " [dev/ino: %d/%llu]",
-				(int) buf->st_dev, (unsigned long long) buf->st_ino);
+
+	if (pseudo_util_debug_flags & PDBGF_OP) {
+		pseudo_debug(PDBGF_OP, "%s%s", pseudo_op_name(op),
+			(dirfd != -1 && dirfd != AT_FDCWD && op != OP_DUP) ? "at" : "");
+		if (path_extra_1) {
+			pseudo_debug(PDBGF_OP, " %s ->", path_extra_1);
 		}
-		pseudo_debug(PDBGF_OP, " (0%o)", (int) buf->st_mode);
+		if (path) {
+			pseudo_debug(PDBGF_OP, " %s", path);
+		}
+		/* for OP_RENAME in renameat, "fd" is also used for the
+		 * second dirfd.
+		 */
+		if (fd != -1 && op != OP_RENAME) {
+			pseudo_debug(PDBGF_OP, " [fd %d]", fd);
+		}
+		if (buf) {
+			pseudo_debug(PDBGF_OP, " (+buf)");
+			if (fd != -1) {
+				pseudo_debug(PDBGF_OP, " [dev/ino: %d/%llu]",
+					(int) buf->st_dev, (unsigned long long) buf->st_ino);
+			}
+			pseudo_debug(PDBGF_OP, " (0%o)", (int) buf->st_mode);
+		}
+		pseudo_debug(PDBGF_OP, ": ");
 	}
-	pseudo_debug(PDBGF_OP, ": ");
 	msg.type = PSEUDO_MSG_OP;
 	msg.op = op;
 	msg.fd = fd;
@@ -1335,7 +1699,8 @@ pseudo_client_op(pseudo_op_t op, int access, int fd, int dirfd, const char *path
 		break;
 	case OP_OPEN:
 		pseudo_client_path(fd, path);
-		do_request = 1;
+	case OP_EXEC: /* fallthrough */
+		do_request = pseudo_client_logging;
 		break;
 	case OP_CLOSE:
 		/* no request needed */
@@ -1391,7 +1756,6 @@ pseudo_client_op(pseudo_op_t op, int access, int fd, int dirfd, const char *path
 		 * (operations which can create should be CREAT or MKNOD
 		 * or MKDIR)
 		 */
-	case OP_EXEC:
 	case OP_CHOWN:
 	case OP_FCHOWN:
 	case OP_FSTAT:
@@ -1412,29 +1776,31 @@ pseudo_client_op(pseudo_op_t op, int access, int fd, int dirfd, const char *path
 		pseudo_diag("error: unknown or unimplemented operator %d (%s)", op, pseudo_op_name(op));
 		break;
 	}
-	if (do_request) {
-		struct timeval tv1, tv2;
+	/* result can only be set when PSEUDO_XATTRDB resulted in a
+	 * successful store to or read from the local database.
+	 */
+	if (do_request && !result) {
+#ifdef PSEUDO_PROFILING
+		struct timeval tv1_ipc, tv2_ipc;
+#endif
                 if (!pseudo_op_wait(msg.op))
                         msg.type = PSEUDO_MSG_FASTOP;
 		pseudo_debug(PDBGF_CLIENT | PDBGF_VERBOSE, "sending request [ino %llu]\n", (unsigned long long) msg.ino);
-		gettimeofday(&tv1, NULL);
+#ifdef PSEUDO_PROFILING
+		gettimeofday(&tv1_ipc, NULL);
+#endif
 		if (pseudo_local_only) {
 			/* disable server */
 			result = NULL;
 		} else {
 			result = pseudo_client_request(&msg, pathlen, path);
 		}
-		gettimeofday(&tv2, NULL);
-		++messages;
-		message_time.tv_sec += (tv2.tv_sec - tv1.tv_sec);
-		message_time.tv_usec += (tv2.tv_usec - tv1.tv_usec);
-		if (message_time.tv_usec < 0) {
-			message_time.tv_usec += 1000000;
-			--message_time.tv_sec;
-		} else while (message_time.tv_usec > 1000000) {
-			message_time.tv_usec -= 1000000;
-			++message_time.tv_sec;
-		}
+#ifdef PSEUDO_PROFILING
+		gettimeofday(&tv2_ipc, NULL);
+		++profile_data.messages;
+		profile_data.ipc_time.tv_sec += (tv2_ipc.tv_sec - tv1_ipc.tv_sec);
+		profile_data.ipc_time.tv_usec += (tv2_ipc.tv_usec - tv1_ipc.tv_usec);
+#endif
 		if (result) {
 			pseudo_debug(PDBGF_OP, "(%d) %s", getpid(), pseudo_res_name(result->result));
 			if (op == OP_STAT || op == OP_FSTAT) {
@@ -1450,20 +1816,30 @@ pseudo_client_op(pseudo_op_t op, int access, int fd, int dirfd, const char *path
 					(int) result->uid,
 					(int) result->gid);
 			}
-		} else {
+		} else if (msg.type != PSEUDO_MSG_FASTOP) {
 			pseudo_debug(PDBGF_OP, "(%d) no answer", getpid());
 		}
-	} else {
+	} else if (!result) {
 		pseudo_debug(PDBGF_OP, "(%d) (no request)", getpid());
 	}
+	#ifdef PSEUDO_XATTRDB
+	else {
+		pseudo_debug(PDBGF_OP, "(%d) (handled through xattrdb: %s)", getpid(), pseudo_res_name(result->result));
+	}
+	#endif
 	pseudo_debug(PDBGF_OP, "\n");
 
-	if (do_request && (messages % 1000 == 0)) {
-		pseudo_debug(PDBGF_CLIENT | PDBGF_VERBOSE | PDBGF_BENCHMARK, "%d messages handled in %.4f seconds\n",
-			messages,
-			(double) message_time.tv_sec +
-			(double) message_time.tv_usec / 1000000.0);
+#ifdef PSEUDO_PROFILING
+	gettimeofday(&tv2_op, NULL);
+	profile_data.op_time.tv_sec += (tv2_op.tv_sec - tv1_op.tv_sec);
+	profile_data.op_time.tv_usec += (tv2_op.tv_usec - tv1_op.tv_usec);
+	if (profile_data.total_ops % profile_interval == 0) {
+		pseudo_profile_report();
+		if (profile_interval < 100) {
+			profile_interval = profile_interval * 10;
+		}
 	}
+#endif
 	/* reenable wrappers */
 	pseudo_magic();
 
@@ -1542,7 +1918,7 @@ populate_path_segs(void) {
 	path_lens[c] = 0;
 }
 
-char *
+const char *
 pseudo_exec_path(const char *filename, int search_path) {
 	char *path = getenv("PATH");
 	char *candidate;
@@ -1599,18 +1975,15 @@ pseudo_exec_path(const char *filename, int search_path) {
 				pseudo_diag("couldn't allocate intermediate path.\n");
 				candidate = NULL;
 			}
-			free(dir);
 		}
 		if (candidate && !stat(candidate, &buf) && !S_ISDIR(buf.st_mode) && (buf.st_mode & 0111)) {
 			pseudo_debug(PDBGF_CLIENT | PDBGF_VERBOSE, "exec_path: %s => %s\n", filename, candidate);
 			pseudo_magic();
 			return candidate;
-		} else {
-			free(candidate);
 		}
 	}
 	/* blind guess being as good as anything */
 	pseudo_magic();
-	return strdup(filename);
+	return filename;
 }
 
