@@ -780,7 +780,7 @@ pdb_log_entry(log_entry *e) {
 		sqlite3_bind_int(insert, field++, e->client);
 		sqlite3_bind_int(insert, field++, e->dev);
 		sqlite3_bind_int(insert, field++, e->gid);
-		sqlite3_bind_int(insert, field++, e->ino);
+		sqlite3_bind_int64(insert, field++, e->ino);
 		sqlite3_bind_int(insert, field++, e->mode);
 		if (e->path) {
 			sqlite3_bind_text(insert, field++, e->path, -1, SQLITE_STATIC);
@@ -874,7 +874,7 @@ pdb_log_msg(pseudo_sev_t severity, pseudo_msg_t *msg, const char *program, const
 		sqlite3_bind_int(insert, field++, msg->client);
 		sqlite3_bind_int(insert, field++, msg->dev);
 		sqlite3_bind_int(insert, field++, msg->gid);
-		sqlite3_bind_int(insert, field++, msg->ino);
+		sqlite3_bind_int64(insert, field++, msg->ino);
 		sqlite3_bind_int(insert, field++, msg->mode);
 		if (msg->pathlen) {
 			sqlite3_bind_text(insert, field++, msg->path, -1, SQLITE_STATIC);
@@ -1175,7 +1175,7 @@ pdb_query(char *stmt_type, pseudo_query_t *traits, unsigned long fields, int uni
 		case PSQF_STAMP:
 		case PSQF_TYPE:
 		case PSQF_UID:
-			sqlite3_bind_int(stmt, field++, trait->data.ivalue);
+			sqlite3_bind_int64(stmt, field++, trait->data.ivalue);
 			break;
 		default:
 			pseudo_diag("Inexplicably invalid field type %d\n", trait->field);
@@ -1396,12 +1396,12 @@ pdb_link_file(pseudo_msg_t *msg, long long *row) {
 		sqlite3_bind_text(insert, 1, "NAMELESS FILE", -1, SQLITE_STATIC);
 	}
 	sqlite3_bind_int(insert, 2, msg->dev);
-	sqlite3_bind_int(insert, 3, msg->ino);
+	sqlite3_bind_int64(insert, 3, msg->ino);
 	sqlite3_bind_int(insert, 4, msg->uid);
 	sqlite3_bind_int(insert, 5, msg->gid);
 	sqlite3_bind_int(insert, 6, msg->mode);
 	sqlite3_bind_int(insert, 7, msg->rdev);
-	pseudo_debug(PDBGF_DB, "linking %s: dev %llu, ino %llu, mode %o, owner %d\n",
+	pseudo_debug(PDBGF_DB | PDBGF_FILE, "linking %s: dev %llu, ino %llu, mode %o, owner %d\n",
 		(msg->pathlen ? msg->path : "<nil> (as NAMELESS FILE)"),
 		(unsigned long long) msg->dev, (unsigned long long) msg->ino,
 		(int) msg->mode, msg->uid);
@@ -1441,7 +1441,7 @@ pdb_unlink_file_dev(pseudo_msg_t *msg) {
 		return 1;
 	}
 	sqlite3_bind_int(sql_delete, 1, msg->dev);
-	sqlite3_bind_int(sql_delete, 2, msg->ino);
+	sqlite3_bind_int64(sql_delete, 2, msg->ino);
 	file_db_dirty = 1;
 	rc = sqlite3_step(sql_delete);
 	if (rc != SQLITE_DONE) {
@@ -1477,7 +1477,7 @@ pdb_update_file_path(pseudo_msg_t *msg) {
 	}
 	sqlite3_bind_text(update, 1, msg->path, -1, SQLITE_STATIC);
 	sqlite3_bind_int(update, 2, msg->dev);
-	sqlite3_bind_int(update, 3, msg->ino);
+	sqlite3_bind_int64(update, 3, msg->ino);
 	file_db_dirty = 1;
 	rc = sqlite3_step(update);
 	if (rc != SQLITE_DONE) {
@@ -1880,7 +1880,7 @@ pdb_update_inode(pseudo_msg_t *msg) {
 		return 1;
 	}
 	sqlite3_bind_int(update, 1, msg->dev);
-	sqlite3_bind_int(update, 2, msg->ino);
+	sqlite3_bind_int64(update, 2, msg->ino);
 	rc = sqlite3_bind_text(update, 3, msg->path, -1, SQLITE_STATIC);
 	if (rc) {
 		/* msg->path can never be null, and if msg didn't
@@ -1933,7 +1933,7 @@ pdb_update_file(pseudo_msg_t *msg) {
 	sqlite3_bind_int(update, 3, msg->mode);
 	sqlite3_bind_int(update, 4, msg->rdev);
 	sqlite3_bind_int(update, 5, msg->dev);
-	sqlite3_bind_int(update, 6, msg->ino);
+	sqlite3_bind_int64(update, 6, msg->ino);
 
 	file_db_dirty = 1;
 	rc = sqlite3_step(update);
@@ -1974,7 +1974,7 @@ pdb_find_file_exact(pseudo_msg_t *msg, long long *row) {
 		dberr(file_db, "error binding %s to select", msg->pathlen ? msg->path : "<nil>");
 	}
 	sqlite3_bind_int(select, 2, msg->dev);
-	sqlite3_bind_int(select, 3, msg->ino);
+	sqlite3_bind_int64(select, 3, msg->ino);
 	rc = sqlite3_step(select);
 	switch (rc) {
 	case SQLITE_ROW:
@@ -2084,7 +2084,7 @@ pdb_get_file_path(pseudo_msg_t *msg) {
 		return 0;
 	}
 	sqlite3_bind_int(select, 1, msg->dev);
-	sqlite3_bind_int(select, 2, msg->ino);
+	sqlite3_bind_int64(select, 2, msg->ino);
 	rc = sqlite3_step(select);
 	switch (rc) {
 	case SQLITE_ROW:
@@ -2113,7 +2113,7 @@ pdb_get_file_path(pseudo_msg_t *msg) {
 
 /* find file using dev/inode as key */
 int
-pdb_find_file_dev(pseudo_msg_t *msg, long long *row) {
+pdb_find_file_dev(pseudo_msg_t *msg, long long *row, char **path) {
 	static sqlite3_stmt *select;
 	int rc;
 	char *sql = "SELECT * FROM files WHERE dev = ? AND ino = ?;";
@@ -2133,7 +2133,7 @@ pdb_find_file_dev(pseudo_msg_t *msg, long long *row) {
 		return 1;
 	}
 	sqlite3_bind_int(select, 1, msg->dev);
-	sqlite3_bind_int(select, 2, msg->ino);
+	sqlite3_bind_int64(select, 2, msg->ino);
 	rc = sqlite3_step(select);
 	switch (rc) {
 	case SQLITE_ROW:
@@ -2145,6 +2145,12 @@ pdb_find_file_dev(pseudo_msg_t *msg, long long *row) {
 		msg->mode = (unsigned long) sqlite3_column_int64(select, 6);
 		msg->rdev = (unsigned long) sqlite3_column_int64(select, 7);
 		msg->deleting = (int) sqlite3_column_int64(select, 8);
+		/* stash path */
+		if (path) {
+			*path = strdup((char *) sqlite3_column_text(select, 1));
+			pseudo_debug(PDBGF_FILE, "find_file_dev: path %s\n",
+				*path ? *path : "<nil>");
+		}
 		rc = 0;
 		break;
 	case SQLITE_DONE:
@@ -2229,7 +2235,7 @@ pdb_list_xattr(long long file_id, char **value, size_t *len) {
 	size_t used = 0;
 	char *buffer = 0;
 	int rc;
-	char *sql = "SELECT name FROM xattrs WHERE file_id = ?;";
+	char *sql = "SELECT name FROM xattrs WHERE file_id = ? ORDER BY name;";
 
 	/* if we don't have a record of the file, it must not have
 	 * any extended attributes...
@@ -2459,7 +2465,7 @@ pdb_find_file_ino(pseudo_msg_t *msg, long long *row) {
 	if (!msg) {
 		return 1;
 	}
-	sqlite3_bind_int(select, 1, msg->ino);
+	sqlite3_bind_int64(select, 1, msg->ino);
 	rc = sqlite3_step(select);
 	switch (rc) {
 	case SQLITE_ROW:
